/*
 * Server binding/registration.
 *
 * Copyright (C) 2018, 2019, 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>

#include <ipc/message.h>

EXTERN_C_BEGIN

/* A handler function type. */

typedef void (*ipc_server_handler_type)(ipc_message_t *, void *);

/* A finaliser function type. */

struct ipc_server_config_type;

typedef void (*ipc_server_finaliser_type)(struct ipc_server_config_type *);

/* A server configuration type. */

typedef struct ipc_server_config_type
{
  /* Component object references. C++ objects may provide different references
     for the handler (providing access to the exposed interface) and the
     finaliser (providing access to a distinct control interface). */

  void *handler_obj;
  void *finaliser_obj;

  /* Expected capability item limit for incoming messages. */

  int expected_items;

  /* Handler for incoming messages and finaliser upon deletion. */

  ipc_server_handler_type handler;
  ipc_server_finaliser_type finaliser;

  /* Create a new thread. */

  int new_thread;

  /* Thread and IPC capabilities. */

  l4_cap_idx_t thread;
  l4_cap_idx_t server;

  /* Receive deletion notifications via an IRQ. */

  int notifications;
  l4_cap_idx_t irq;

} ipc_server_config_type;



/* Associate a notification IRQ with an IPC gate in the main thread. */

long ipc_server_apply_irq(l4_cap_idx_t cap, l4_cap_idx_t *irq);

/* Associate a notification IRQ with an IPC gate in the given thread. */

long ipc_server_apply_irq_for_thread(l4_cap_idx_t cap, l4_cap_idx_t *irq, l4_cap_idx_t thread);

/* Bind the main thread to a named IPC gate capability. */

long ipc_server_bind(const char *name, l4_umword_t id, l4_cap_idx_t *server);

/* Create a new IPC gate for the main thread. */

long ipc_server_new(l4_cap_idx_t *cap, void *obj);

/* Create a new IPC gate for the given thread. */

long ipc_server_new_for_thread(l4_cap_idx_t *cap, void *obj, l4_cap_idx_t thread);

/* Create an IPC gate for the main thread. */

long ipc_server_new_gate(l4_cap_idx_t *ref, l4_umword_t id);

/* Create an IPC gate for the indicated thread. */

long ipc_server_new_gate_for_thread(l4_cap_idx_t *ref, l4_cap_idx_t thread, l4_umword_t id);



/* Handle incoming messages for a server. */

long ipc_server_loop(int expected_items, ipc_server_config_type *config);

/* Handle incoming messages and IRQ notifications for a server. */

long ipc_server_managed_loop(int expected_items, ipc_server_config_type *config,
                             l4_cap_idx_t irq);

/* A pthread-compatible mainloop initiation function. */

void *ipc_server_start_mainloop(void *data);

/* Wait for an incoming message via a dedicated IPC gate for a thread. */

l4_msgtag_t ipc_server_wait(l4_umword_t id);



/* Finalise a server configuration. */

void ipc_server_finalise_config(ipc_server_config_type *config);

/* Discard any thread created but not initiated. */

void ipc_server_discard_thread(ipc_server_config_type *config);

/* Initialise a server configuration. */

void ipc_server_init_config(ipc_server_config_type *config);

/* Initialise and start a server using the given configuration. */

long ipc_server_start_config(ipc_server_config_type *config);

EXTERN_C_END
